#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <dirent.h>
#include <limits.h>
#include "vfs.h"
#include <remaudio/dispatch.h>

/*
	Status qui seront rutilis dans server.c
*/
char vopen_copyonwrite;
char vopen_trunconwrite;


/*
	Ouverture d'un fichier dans un rpertoire virtuel.
	(voir les fonction vdir_xxxx)

	Vrifie si le path fname correspond existe localement ou dans
	l'archive.

	Si le fichier est ouvert en criture et que le fichier est prsent
	dans l'archive et pas localement, le fichier sera copier localement
	avant l'ouverture.

	Retourne NULL si erreur ou ne peut ouvrir ou fichier existe pas.

	Attention: L'ouverture en criture seulement est identique 
		open (un peu plus lent peut-tre).
*/
static int vopen (
	const char *fname,
	int   mode,
	int acces)		/* Sert seulement avec O_CREAT */
{
	#if 0
	{
		char tmp[1000];
		int len = sprintf (tmp,"open :%s:\n",fname);
		write (2,tmp,len);
	}
	#endif
	char reel[PATH_MAX];
	int exist = vdir_exist (fname,reel);
	#if 0
	{
		char tmp[1000];
		int len = sprintf (tmp,"reel %d :%s:\n",exist,reel);
		write (2,tmp,len);
	}
	#endif
	vopen_copyonwrite = 0;
	vopen_trunconwrite = 0;
	if (exist == 2){
		/*
			Le fichier est prsent dans l'archive.
			Si ouverture pour modification (pas seulement criture),
			on doit copier le fichier
		*/
		fname = reel;
		if ((mode & O_APPEND) != 0
			|| (mode & O_RDWR)==O_RDWR){
			vopen_copyonwrite = 1;
		}else if ((mode & (O_WRONLY|O_TRUNC|O_CREAT))==0){
			/* Forcement mode lecture */
			vopen_copyonwrite = 1;
		}else{
			vopen_copyonwrite = 1;
			vopen_trunconwrite = 1;
		}
		mode = O_RDONLY;
	}
	return __libc_open (fname,mode,acces);
}


static int vclose (int fd)
{
	return __libc_close (fd);
}

static int vioctl (int fd, unsigned request, void *a1, void *a2, void *a3)
{
	return __libc_ioctl (fd,request,a1,a2,a3);
}


static int vstat (int ver, const char *fname, struct stat *buf)
{
	char reel[PATH_MAX];
	int exist = vdir_exist (fname,reel);
	#if 0
	{
		char tmp[1000];
		int len = sprintf (tmp,"vstat :%s: -> %d :%s:\n",fname,exist,reel);
		write (2,tmp,len);
	}
	#endif
	#if 1
		int ret = DISPATCH_DONTCARE;
		if (exist == 2){
			ret = __xstat (ver,reel,buf);
		}
		return ret;
	#else
		if (exist == 2) fname = reel;
		return __xstat (ver,fname,buf);
	#endif
}
static int vlstat (int ver, const char *fname, struct stat *buf)
{
	char reel[PATH_MAX];
	int exist = vdir_exist (fname,reel);
	#if 0
	{
		char tmp[1000];
		int len = sprintf (tmp,"vlstat :%s: -> %d :%s:\n",fname,exist,reel);
		write (2,tmp,len);
	}
	#endif
	int ret = DISPATCH_DONTCARE;
	if (exist == 2){
		ret = __lxstat (ver,reel,buf);
	}
	return ret;
}

static int vunlink (const char *fname)
{
	int ret = 0;
	char name[MAXSIZ_NAME];
	char path[MAXSIZ_PATH];
	char makefile_dat[MAXSIZ_PATH];
	path_splitlex (fname,path,name);
	path_make (path,"makefile.dat",makefile_dat);
	if (!vfile_exist (makefile_dat)){
		ret = unlink (fname);
	}else if (strcmp(name,"makefile.dat")==0
		|| strcmp(name,"makefile.del")==0){
		ret = -1;
	}else{
		int exist = vdir_exist(fname,NULL);
		if (exist == 0){
			ret = -1;
		}else{
			if (exist == 1){
				/*
					le fichier est prsent localement dans le
					rpertoire.
				*/
				ret = unlink (fname);
			}
			if (ret != -1 && vdir_unlink (fname) != -1){
				FILE *fout;
				/* Calcul le path du fichier makefile.del */
				char pathdel[MAXSIZ_PATH];
				path_make (path,"makefile.del",pathdel);
				fout = fopen (pathdel,"a");
				if (fout != NULL){
					fprintf (fout,"%s\n",name);
					ret = fclose (fout);
				}else{
					ret = -1;
				}
			}
		}
	}
	return ret;
}

static int vrename (const char *from, const char *to)
{
	int ret = 0;
	char name[MAXSIZ_NAME];
	char path[MAXSIZ_PATH];
	char makefile_dat[MAXSIZ_PATH];
	path_splitlex (from,path,name);
	path_make (path,"makefile.dat",makefile_dat);
	if (!vfile_exist (makefile_dat)){
		ret = rename (from,to);
	}else if (strcmp(name,"makefile.dat")==0
		|| strcmp(name,"makefile.del")==0){
		ret = -1;
	}else{
		char realpath[MAXSIZ_PATH];
		int exist = vdir_exist(from,realpath);
		if (exist == 0){
			ret = -1;
		}else if (exist == 1){
			/*
				le fichier est prsent localement dans le
				rpertoire.
			*/
			ret = rename (from,to);
			if (ret == 0){
				vunlink (from);
			}						
		}else if (exist == 2){
			/*
				Le fichier est dans /kit/ombre, on doit faire un copy
			*/
			ret = file_copyp (realpath,to) == 1 ? 0 : -1;
			if (ret == 0){
				vunlink (from);
			}
		}
	}
	return ret;
}

struct VIRT_DIR{
	int nb;
	char **tb;
	int pos;
	DIR *dir;
};


static DIR *vopendir(const char *dname)
{
	DIR *ret = NULL;
	char makefile_dat[PATH_MAX];
	snprintf (makefile_dat,sizeof(makefile_dat)-1,"%s/makefile.dat",dname);
	struct stat st;
	if (stat(makefile_dat,&st)!=-1){
		VIRT_DIR *vdir = (VIRT_DIR*)calloc (1,sizeof(VIRT_DIR));
		if (vdir != NULL){
			char *tb[2000];
			vdir->nb = vdir_getlist (dname,"*",tb,2000);
			vdir->nb += vdir_getlistd (dname,"*",tb+vdir->nb,2000-vdir->nb);
			//DEBUG (VDBG,("vdir_getlist :%s: %d\n",dname,vdir->nb));
			if (vdir->nb > 0){
				int size = vdir->nb*sizeof(char*);
				vdir->tb = (char**)malloc(size);
				if (vdir->tb == NULL){
					vdir->nb = 0;
				}else{
					memcpy (vdir->tb,tb,size);
					ret = (DIR*) vdir;
				}
			}
		}
	}
	return ret;
}

static int vclosedir(DIR *dir)
{
	int ret = -1;
	if (dir != NULL){
		VIRT_DIR *vdir = (VIRT_DIR*)dir;
		ret = 0;
		if (vdir->dir != NULL) ret = __closedir (vdir->dir);
		tbstr_free (vdir->tb,vdir->nb);
		free (vdir->tb);
		free (vdir);
	}
	return ret;
}

struct dirent *vreaddir(DIR *dir)
{
	struct dirent *ret = NULL;
	if (dir != NULL){
		VIRT_DIR *vdir = (VIRT_DIR*)dir;
		while (1){
			if (vdir->pos < vdir->nb){
				static struct dirent dd;
				path_splitlex (vdir->tb[vdir->pos],NULL,dd.d_name);
				vdir->pos ++;
				if (strcmp(dd.d_name,"makefile.dat")!=0
					&& strcmp(dd.d_name,"makefile.del")!=0){
					static int alloc_ino = 1;
					dd.d_ino = alloc_ino++;
					ret = &dd;
					break;
				}
			}else{
				if (vdir->dir != NULL){
					ret = __readdir (vdir->dir);
				}
				break;
			}
		}
	}
	return ret;
}

static DISPATCH m={
	NULL,
	vopen,
	vclose,
	__libc_write,
	__libc_read,
	vioctl,
	__libc_lseek,
	vstat,
	__fxstat,
	vlstat,
	__chdir,
	vopendir,
	vreaddir,
	vclosedir,
	vrename,
	vunlink,
	NULL,		// rmdir
	NULL,		// mkdir
	NULL,		// socket
	NULL,		// connect
	NULL,		// bind
};

static REGISTER_DISPATCH mm (m);

